@title "Records"
begin
record state(
   reference(transition) out;
      comment a TNEXT list of all transitions leading from
         this state;
   reference(transition) in;
      comment a single transition leading to this state.
         Following 'IN' will eventually lead to the start state;
   reference(conflist) conf;
   reference(state) snext;
      comment the next state in a linked list of all
         states accessed by the same symbol;
   reference(state) slink;
      comment 'SLINK' is used by 'GROWFSM' and 'MAKESTATE';
   integer statenumber
   );

record transition(
   reference(state) from;
   reference(symbol, production) tsymbol;
   reference(state) to;
   reference(transition) tnext;
   integer transitionnumber
   );

record production(
   reference(symbol) left; comment non-terminal on left;
   reference(configuration) right; comment string on right;
   reference(production) pnext;
   integer productionnumber
   );

record symbol(
   string(8) name;
   integer symnum;
   reference(production) definitions;
      comment productions with non-terminal on left;
   reference(symbol) nextsymbol;
   logical isterminal;
   reference(conflist) attachment;
   reference(state) states
      comment states accessed by the symbol;
   );

record configuration(
   reference(symbol) sym;
   reference(configuration, production) promote
   );

record conflist(
   reference(configuration) this;
   reference(conflist) rest
   );
@title "Read grammar"
reference(symbol) symbols, terminals, nonterminals;
integer nterms, lownt, highnt, nsymbols, nproductions;
reference(production) startproduction;
reference(symbol) superstart, endoffile;

procedure readgrammar;
   begin
   string(256) line, blank;
   integer pn, j;
   reference(configuration, production) c;
   reference(production) p;
   symbols := terminals := nonterminals := null;
   readcard(blank);
   write(blank);
   readcard(line);
   superstart := findsymbol("* start ");
   endoffile := findsymbol("end file");

   pn := 0; startproduction := null;
   while line ~= blank
   do begin
      pn := pn + 1;
      write(pn, line);
      p := production(null, null, null, pn);
      j := ((256 - 10) div 9) * 9 + 11 - 9;
      for i := 11 step 9 until j
      do if line(i|8) = "        "
         then begin j := i - 9; go to endloop end;
endloop:
      c := p;
      for i := j step -9 until 11
      do c := configuration(findsymbol(line(i|8)), c);
      right(p) := c;
      left(p) := findsymbol(line(0|8));
      pnext(p) := definitions(left(p));
      definitions(left(p)) := p;
      isterminal(left(p)) := false;
      if definitions(superstart) = null
      then begin
         startproduction := production(null, null, null, 0);
         left(startproduction) := superstart;
         right(startproduction) :=
            configuration(
               left(p),
               configuration(endoffile, startproduction)
            );
         definitions(superstart) := startproduction;
         isterminal(superstart) := false
         end;
      readcard(line)
      end;
   write(line);
   nproductions := pn
   end readgrammar;

reference(symbol) procedure findsymbol(string(8) value s);
   begin
   reference(symbol) t;
   t := symbols;
   while (t ~= null) and (name(t) ~= s)
   do t := nextsymbol(t);
   if t = null
   then t := symbols :=
         symbol(s, 0, null, symbols, true, null, null);
   t
   end findsymbol;

procedure sortsymbols;
   begin
   reference(symbol) s, t, n, u;
   integer i;
   t := n := null;
   s := symbols;
   while s ~= null
   do begin
      u := nextsymbol(s);
      if isterminal(s)
      then begin nextsymbol(s) := t; t := s end
      else begin nextsymbol(s) := n; n := s end;
      s := u;
      end;
   if t = null
   then begin
      nterms := 0; lownt := 1; i := 0;
      s := n;
      while s ~= null
      do begin
         i := i + 1;
         symnum(s) := i;
         s := nextsymbol(s)
         end
      end
   else begin
      s := t; u := null; i := 0;
      while (s ~= null) and (nextsymbol(s) ~= null)
      do begin
         i := i + 1;
         symnum(s) := i;
         s := nextsymbol(s)
         end;
      nextsymbol(s) := n;
      nterms := i + 1;
      lownt := nterms + 1;
      while s ~= null
      do begin
         i := i + 1;
         symnum(s) := i;
         s := nextsymbol(s)
         end
      end;
   symbols := terminals := t;
   nonterminals := n;
   highnt := i;
   nsymbols := i
   end sortsymbols;

procedure printsymbols;
   begin
   reference(symbol) s;
   s := symbols;
   write("Terminals");
   while s ~= null
   do begin
      if s = nonterminals
      then write("Non-terminals ");
      write(symnum(s), name(s));
      s := nextsymbol(s)
      end
   end;

integer nsymbols32, lownt32, highnt32;

write("Time: ", time(2) / 38400, " seconds.");
readgrammar;
write("Time: ", time(2) / 38400, " seconds.");
sortsymbols;
write("Time: ", time(2) / 38400, " seconds.");
printsymbols;
write("Time: ", time(2) / 38400, " seconds.");

nsymbols32 := nsymbols div 32;
lownt32 := lownt div 32;
highnt32 := highnt div 32;

begin

reference(symbol) array symbolarray(1 :: nsymbols);

procedure makesymbolarray;
   begin
   reference(symbol) s;
   s := symbols;
   while s ~= null
   do begin
      symbolarray(symnum(s)) := s;
      s := nextsymbol(s)
      end
   end makesymbolarray;
@title "Bit arrays"
bits array plnonempty(lownt32 :: highnt32, 1 :: nsymbols);
bits(32) array f(0 :: nsymbols32, 1 :: nsymbols);

logical procedure bit(
      bits array a(*, *);
      integer value i, j);
   begin
   bits element;
   integer shift;
   element := a(i div 32, j);
   shift := i rem 32;
   (element shl shift shr 31) = #1
   end bit;

procedure setbit(
      bits array a(*, *);
      integer value i, j);
   begin
   integer idiv32;
   idiv32 := i div 32;
   a(idiv32, j) := a(idiv32, j) or
      (#80000000 shr (i rem 32))
   end setbit;

procedure createbitarrays;
   comment compute F and PLNONEMPTY. The desired arrays
      satasfy
         PL(i, j) iff i =>* j... .
         PR(i, j) iff i =>* ...j .
         PLNONEMPTY(i, j) iff i =>* j... using no productions
            of the form A => empty.
         F(i, j) if S =>* ...ij... ;
   begin
   logical array empty(1 :: nsymbols);
   bits(32) array pr, pl(lownt32 :: highnt32, 1 :: nsymbols);
   bits(32) array adjt, temp(0 :: nsymbols32, 1 :: nsymbols);
@title "Bit arrays"
   comment find which nonterminals can produce the empty string;

      begin
      reference(symbol) n;
      reference(production) p;
      reference(configuration) c;
      logical anye, e;
      logical repeat;

      for i := 1 until nsymbols
      do empty(i) := false;

      for i := lownt32 until highnt32
      do for j := 1 until nsymbols
         do pr(i, j) := pl(i, j) := plnonempty(i, j) := #0;

      for i := 0 until nsymbols32
      do for j := 1 until nsymbols
         do adjt(i, j) := temp(i, j) := f(i, j) := #0;

      repeat := true;
      while repeat
      do begin
         repeat := false;
         n := nonterminals;
         while n ~= null
         do begin
            if ~ empty(symnum(n))
            then begin
               p := definitions(n);
               anye := false;
               while (p ~= null) and ~ anye
               do begin
                  c := right(p);
                  e := true;
                  while (c is configuration) and e
                  do begin
                     e := e and empty(symnum(sym(c)));
                     c := promote(c)
                     end;
                  anye := anye or e;
                  p := pnext(p)
                  end;
               if anye and ~ empty(symnum(n))
               then begin
                  repeat := true;
                  empty(symnum(n)) := true
                  end
               end;
            n := nextsymbol(n)
            end
         end

      end;

   comment At this point, EMPTY(i) iff i =>* empty;

      begin reference(symbol) n;
      write(" ");
      write("The empty string can be derived from the ",
         "following non-terminals: ");
      n := nonterminals;
      while n ~= null
      do begin
         if empty(symnum(n))
         then write(name(n));
         n := nextsymbol(n)
         end
      end;
@title "Bit arrays"
   comment compute PL, PR, and ADJ.
         PL(i, j) iff i => xxxj... and xxx =>* empty.
         PR(i, j) iff i => ...jxxx and xxx =>* empty.
         ADJ(i, j) IFF x => ...ixxxj... and xxx =>* empty;

      begin
      reference(symbol) n;
      reference(production) p;
      reference(configuration, production) c, cf, prev, ctest;
      n := nonterminals;
      while n ~= null
      do begin
         p := definitions(n);
         while p ~= null
         do begin
            c := right(p);
            if c is configuration
            then begin
               setbit(plnonempty, symnum(n), symnum(sym(c)));
               prev := p;
               while c ~= null
               do begin
                  cf := c;
                  while (begin
                     if prev is production
                     then begin
                        if cf is production
                        then
                        else setbit(pl, symnum(n),
                              symnum(sym(cf)))
                        end
                     else if cf is production
                        then setbit(pr, symnum(n),
                              symnum(sym(prev)))
                        else setbit(adjt, symnum(sym(cf)),
                           symnum(sym(prev)));

                     (cf is configuration)
                     and empty(symnum(sym(cf)))
                     end)
                  do cf := promote(cf);

                  prev := c;
                  c := if c is configuration
                     then promote(c) else null
                  end
               end;
            p := pnext(p)
            end;
         n := nextsymbol(n)
         end
      end;

   comment At this point,
         PL(i, j) iff i => xxxj... and xxx =>* empty.
         PR(i, j) iff i => ...jxxx and xxx =>* empty.
         ADJ(i, j) IFF x => ...ixxxj... and xxx =>* empty;

@title "Bit arrays"
   comment compute the symmetric transitive closures
                    of PL, PR, and PLNONEMPTY;

      for j := lownt until highnt
      do for k := 1 until nsymbols
         do begin
            if bit(pr, j, k)
            then for i := lownt32 until highnt32
               do pr(i, k) := pr(i, k) or pr(i, j);

            if bit(pl, j, k)
            then for i := lownt32 until highnt32
               do pl(i, k) := pl(i, k) or pl(i, j);

            if bit(plnonempty, j, k)
            then for i := lownt32 until highnt32
               do plnonempty(i, k) := plnonempty(i, k)
                  or plnonempty(i, j);
            end;

      for i := lownt until highnt
      do begin
         setbit(pr, i, i);
         setbit(pl, i, i);
         setbit(plnonempty, i, i);
         end;


   comment at this point,
         PL(i, j) iff i =>* j... .
         PR(i, j) iff i =>* ...j .
         PLNONEMPTY(i, j) iff i =>* j... using no productions
            of the form A => empty;

   comment Now compute F, the 'follows' matrix
      F = transpose(PR*) * ADJ * PL*
        = transpose(PR) * (transpose(ADJT) * PL);

   begin
   bits(32) l;

   for i := 1 until nsymbols
   do for j := 1 until nsymbols
      do begin
         l := #0;
         for k := lownt32 until highnt32
         do l := l or (adjt(k, i) and pl(k, j));
         if (l ~= #0) or bit(adjt, j, i)
         then setbit(temp, i, j)
         end;

   for i := 1 until nsymbols
   do for j := 1 until nsymbols
      do begin
         l := #0;
         for k := lownt32 until highnt32
         do l := l or (pr(k, i) and temp(k, j));
         if (l ~= #0) or bit(temp, i, j)
         then setbit(f, i, j)
         end
   end;

   printmatrix(pr, lownt, highnt, 1, nsymbols, "R");
   printmatrix(pl, lownt, highnt, 1, nsymbols, "L");
   printmatrix(plnonempty, lownt, highnt, 1, nsymbols, ">");
   printmatrix(f, 1, nsymbols, 1, nsymbols, "F");

   end createbitarrays;
@title "Print bit matrix"
procedure printmatrix(
   bits(32) array a(*, *);
   integer value i, j, k, l;
   string(1) value mark);
   begin
   string(100) line;
   integer q, s;
   string(2) background;

   for p := i step 50 until j
   do begin
      q := if j > p+49 then p+49 else j;
      for r := k step 100 until l
      do begin
         s := if l > r+99 then r+99 else l;
         iocontrol(3);
         for z := 0 until 7
         do begin
            line := " ";
            for u := r until s
            do line(u - r| 1 ) := name(symbolarray(u))(z|1);
            write("                        ", line)
            end z;
         for t := p until q
         do begin
            background := if t rem 5 = 0 then "-+" else " |";
            line := " ";
            for u := r until s
            do begin
               line(u - r | 1) :=
                  if bit(a, t , u) then mark
                  else if u rem 5 = 0 then background(1|1)
                  else background(0|1)
               end u;
            write(t, name(symbolarray(t)), line)
            end t;
         end r
      end p
   end printarray;
@title "FSM states"
integer nstates, ntransitions;

reference(state) startstate,
                 newstates;

reference(state) procedure makestate(
      reference(conflist) value c;
      reference(symbol) value access);

   begin reference(state) thesestates, s;

   if access = null
   then thesestates := null
      comment 'makestates' is called only once
         with 'access' = 'null';
   else thesestates := states(access);

   comment search for an equivalent state;

   s := thesestates;
   while (s ~= null) and ~ equal(conf(s), c)
   do s := snext(s);

   if s = null

   then begin
      comment make a new state;

      nstates := nstates + 1;
      s := newstates := thesestates :=
         state(null, null, c, thesestates, newstates, nstates);

      end;

   if access = null
   then startstate := thesestates
   else states(access) := thesestates;

   s
   end makestate;
@title "FSM states"
procedure attach(
      reference(production, configuration) value c;
      reference(symbol) value s);
   begin
   reference(conflist) cl;
   cl := attachment(s);
   while (cl ~= null) and (this(cl) ~= c)
   do cl := rest(cl);
   if cl = null
   then attachment(s) := conflist(c, attachment(s))
   end attach;

procedure maketransition(
      reference(state) value froom;
      reference(symbol, production) value s;
      reference(state) value too);
   begin
   reference(transition) t;
   t := out(froom);
   while (t ~= null)
      and ((from(t) ~= froom) or (tsymbol(t) ~= s)
         or (to(t) ~= too))
   do t := tnext(t);
   if t = null
   then begin
      ntransitions := ntransitions + 1;
      out(froom) := transition(froom, s, too,
         out(froom), ntransitions);
      if (too ~= null) and (in(too) = null)
      then in(too) := out(froom)
      end
   end maketransition;

logical procedure equal(
      reference(conflist) value c, d);
   subset(c, d) and subset(d, c);

logical procedure subset(reference(conflist) value c, d);
   begin
   reference(conflist) l, m;
   l := c;
   while (l ~= null)
      and (begin
         m := d;
         while (m ~= null) and (this(l) ~= this(m))
         do m := rest(m);
         m ~= null
         end)
   do l := rest(l);
   l = null
   end subset;
@title "FSM states"
procedure forallstates(
      reference(state) s; comment by name;
      procedure x);
   begin
   reference(symbol) ac;
   reference(state) st;

   s := startstate;
   x;
   ac := symbols;
   while ac ~= null
   do begin
      st := states(ac);
      while st ~= null
      do begin
         s := st;
         x;
         st := snext(st)
         end;
      ac := nextsymbol(ac)
      end;
   end forallstates;
@title "Grow FSM"
procedure growfsm;
   begin
   reference(state) currentstate;

   nstates := ntransitions := 0;
   newstates := null;

   currentstate := makestate(
         conflist(right(startproduction), null),
         null
         );

   newstates := currentstate;
   while newstates ~= null
   do begin

      reference(conflist) cl;
      reference(configuration, production) c, rp;
      reference(symbol) s, t;
      integer sn;
      reference(production) p;

      currentstate := newstates;
      newstates := slink(newstates);

      cl := conf(currentstate);
      while cl ~= null
      do begin
         c := this(cl);
         if c is production
         then comment a reduce transition is wanted;
              maketransition(currentstate, c, null)
         else if c is configuration
         then comment promotion transitions are wanted,
               but they must be sorted by symbol;
            begin
            s := sym(c);
            sn := symnum(s);
            attach(promote(c), s);
            if sn > nterms
            then comment a non-terminal;
               begin
               t := symbols;
               while t ~= null
               do begin
                  if bit(plnonempty, sn, symnum(t))
                  then begin
                     p := definitions(t);
                     while p ~= null
                     do begin rp := right(p);
                        if rp is production
                        then maketransition(currentstate, p, null)
                        else attach(promote(rp), sym(rp));
                        p := pnext(p)
                        end
                     end;
                  t := nextsymbol(t)
                  end
               end
            end;
         cl := rest(cl)
         end;

      comment new states have been considered as sets of
         configurations. Now make them into real states;

      s := symbols;
      while s ~= null
      do begin
         if attachment(s) ~= null
         then begin reference(state) q;
            q := makestate(attachment(s), s);
            maketransition(currentstate, s, q);
            end;
         attachment(s) := null;
         s := nextsymbol(s)
         end
      end;
   end growfsm;
@title "Lookahead"
procedure checklookahead;
   begin

   procedure inadequacy(
         reference(state) value s);
      begin

      procedure outout(reference(state) value s);
         if (s ~= null) and (in(s) ~= null)
         then begin
            outout(from(in(s)));
            write(" ", name(tsymbol(in(s))))
            end;

      reference(transition) t;

      write("An inadequate state can be reached by reading");
      outout(s);
      write("Possible actions are: ");
      t := out(s);
      while t ~= null
      do begin
         if (tsymbol(t) is symbol)
            and (isterminal(tsymbol(t)))
         then write("   Read ", name(tsymbol(t)))
         else if tsymbol(t) is production
         then begin
            write("   Apply production ",
               productionnumber(tsymbol(t)),
               " with lookaheads ");

            for l := symnum(left(tsymbol(t)))
            do for i := 1 until nterms
               do begin
                  if bit(f, l, i)

                  then writeon(" ", name(symbolarray(i)))
                  end
            end;
         t := tnext(t)
         end;
      error := true
      end inadequacy;

   reference(state) s;
   reference(transition) t;
   logical array b, c(1 :: nterms);
   integer l;
   logical error;

   error := false;
   forallstates(s,
      begin
      t := out(s);
      for i := 1 until nterms
      do b(i) := false;

      while (t ~= null)
         and (comment no conflict;
            if (tsymbol(t) is symbol) and isterminal(tsymbol(t))
            then if b(symnum(tsymbol(t)))
               then false
               else (begin
                  b(symnum(tsymbol(t))) := true;
                  true
                  end)
            else if tsymbol(t) is production
            then (begin
               integer l;
               logical conflict;
               l := symnum(left(tsymbol(t)));
               conflict := false;
               for i := 1 until nterms
               do if b(i) and bit(f, l, i)
                  then conflict := true
                  else b(i) := b(i) or bit(f, l, i);
               ~ conflict
               end)
            else true
            )
      do t := tnext(t);

      if t ~= null
      then inadequacy(s);
      end)
   end checklookahead;
@title "Punch integers"
string(80) l;
integer lpt;

procedure startout;
   begin
   l := " ";
   lpt := -1
   end;

procedure endout;
   if lpt >= 0
   then write(l);

procedure outinteger(integer value i);
   begin
   if (lpt <= 0) or (l(lpt|1) ~= " ")
   then outchar(" ");
   if lpt >= 66
   then begin endout; startout end;
   outint(i)
   end outinteger;

procedure outint(
      integer value i);
   if i < 0
   then begin
      outchar("-");
      outint(- i)
      end
   else if i >= 10
   then begin
      string(10) s;
      outint(i div 10);
      outint(i rem 10)
      end
   else begin
      string(10) s;
      s := "0123456789";
      outchar(s(i|1))
      end;

procedure outchar(string(1) value c);
   begin
   lpt := lpt + 1;
   if lpt >= 72
   then begin
      write(l);
      lpt := 0
      end;
   l(lpt|1) := c
   end outchar;

@title "Punch tables"
procedure punchtables;
   begin

   procedure rowprint(
         integer i;
         integer value from, by, to;
         procedure x);
      begin
      iocontrol(2);
      for j := from step by until to
      do begin
         for k := j step 1 until j + by - 1
         do begin
            i := k;
            if k <= to then x
            end;
         iocontrol(2)
         end;
      write(" ");
      iocontrol(2)

      end rowprint;

   procedure quotedsymbolname(
         string(8) value s);
      begin
      writeon(" """);
      for i := 0 until 7
      do begin
         if s(i|1) = """"
         then writeon("""""")
         else writeon(s(i|1))
         end;
      writeon(""" ")
      end quotedsymbolname;

   integer array a(1 :: nstates, 1 :: nsymbols);
   integer array access(1 :: nstates);
   integer array productionlength,
                 target (0 :: nproductions);

   reference(state) st;
   integer stn;
   reference(transition) tr;
   reference(symbol, production) sy;
   integer oldintfieldsize;
   integer i, j;

   oldintfieldsize := intfieldsize;
   intfieldsize := 7;
   forallstates(st,
      begin
      stn := statenumber(st);
      access(stn) :=
         if (in(st) ~= null)
            and (tsymbol(in(st)) ~= null)
         then symnum(tsymbol(in(st)))
         else 0;

      for i := 1 until nsymbols
      do a(stn, i) := 0;
      tr := out(st);
      while tr ~= null
      do begin
         sy := tsymbol(tr);
         if sy is production
         then begin
            integer pn, l, len;
            pn := productionnumber(sy);
            l := symnum(left(sy));

            target(pn) := l;
            productionlength(pn) :=
               (begin
                integer len;
                reference(configuration, production) v;
                len := 0;
                v := right(sy);
                while v is configuration
                do begin
                   len := len + 1;
                   v := promote(v)
                   end;
                len
                end);

            for i := 1 until nsymbols
            do if bit(f, l, i)
               then a(stn, i) := - pn
            end
         else a(stn, symnum(sy)) := statenumber(to(tr));
         tr := tnext(tr)
         end;
      end);

   iocontrol(3);
   writeon("Machine-readable tables");
   write(nstates, nsymbols, nproductions);

   startout;

   for i := 1 until nstates
   do for j := 1 until nsymbols
      do outinteger(a(i, j));
   for i := 1 until nstates
   do outinteger(access(i));

   endout;

   rowprint(i, 1, 5, nsymbols,
         quotedsymbolname(name(symbolarray(i))));

   startout;

   for i := 1 until nproductions
   do begin
      outinteger(target(i));
      outinteger(productionlength(i))
      end;

   endout;

   write(" "); write(" "); iocontrol(2);


   write(" ");

   intfieldsize := oldintfieldsize
   end punchtables;
@title "Process grammar"
makesymbolarray;
write("Time: ", time(2) / 38400, " seconds.");
createbitarrays;
write("Time: ", time(2) / 38400, " seconds.");
growfsm;
write("Time: ", time(2) / 38400, " seconds.");
checklookahead;
write("Time: ", time(2) / 38400, " seconds.");
punchtables;
write("Time: ", time(2) / 38400, " seconds.");
end
end.
